package com.android.td5;

import android.app.Activity;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.Button;
import android.widget.EditText;
import android.widget.RadioGroup;
import android.widget.RadioGroup.OnCheckedChangeListener;
import android.widget.TextView;

public class Sieve extends Activity {
	// Nested class for continuations passed as messages to the graphical thread
	private class TextViewMessage implements Runnable {
		// TextView to display text into
		private TextView tv;

		// String to update the text of the TextView
		private String s;

		TextViewMessage(TextView tv, String s) {
			this.tv = tv;
			this.s = s;
		}

		@Override
		public void run() {
			// Update the TextView's text
			tv.setText(s);
		}
	}

	// Marking delay menu items
    private static final int MENU_FAST = 0;
    private static final int MENU_SLOW = 1;
    private static final int MENU_PIPELINE = 2;
    private static final int MENU_RACE = 3;
    private static final int MENU_QUIT = 4;
    
    // Keys for the saved state bundle
    private static final String KEY_NUMBER = "number";
    private static final String KEY_DELAY = "delay";
    private static final String KEY_ALGORITHM = "algorithm";
    private static final String KEY_NUM_THREADS = "num_threads";
    private static final String KEY_IS_PRIME = "is_prime";

    // Handler to update graphical elements
	private final Handler callbackHandler = new Handler();

	// Widgets
	private TextView statusTextView;
	private TextView resultTextView;
	private EditText numberEditText;
	
	// Data and computation model of the application
	private SieveModel model;

	// Create an anonymous implementation of OnClickListener
	private OnClickListener startListener = new OnClickListener() {
		@Override
	    public void onClick(View v) {
	    	model.sieve();
		}
	};

	// Create an anonymous implementation of OnCheckedChangeListerner
	private OnCheckedChangeListener numThreadsListener = new OnCheckedChangeListener() {
		@Override
		public void onCheckedChanged(RadioGroup group, int checkedId) {
			int t;
			switch (checkedId) {
			case R.id.radio2:
				t = 2;
				break;
			case R.id.radio4:
				t = 4;
				break;
			default:
				t = 1;
			}
			
			model.setNumThreads(t);
		}
	};

	// Post a message to the graphical thread to set the text of a TextView
	private void print(TextView tv, String s) {
		callbackHandler.post(new TextViewMessage(tv, s));
	}

	// Set the text of the status TextView
	public void printStatus(String s) {
		print(statusTextView, s);
	}
	
	// Set the text of the result TextView
	public void printResult(String s) {
		print(resultTextView, s);
	}

	// Parse the number in the number Edit Text
	public int parseNumber() {
		int n = 0;
		try {
			// Grab the value inside the EditText and check this is an integer value
			n = Integer.parseInt(numberEditText.getText().toString());
			
			// Print the work in progress message
			printStatus("Sieve of n = " + n + " in progress");
		} catch (NumberFormatException e) {
			// Update the TextViewResult
			printResult("Please enter an integer value");
		}

		if (n < 0)
			n = 0;
		if (n > SieveModel.MAX_NUMBER)
			n = SieveModel.MAX_NUMBER;

		return n;
	}

	// Called when the activity is first created
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);
        
        numberEditText = (EditText)findViewById(R.id.numberEditText);
        statusTextView = (TextView)findViewById(R.id.statusTextView);
        resultTextView = (TextView)findViewById(R.id.resultTextView);

        model = new SieveModel(this);
        model.init();
        
        // Capture the SurfaceView to display the sieve array
        SieveSurfaceView sieveSurfaceView = (SieveSurfaceView)findViewById(R.id.sieveSurfaceView);
        // Tell the SurfaceView object where to find the data and algorithms of the application
        sieveSurfaceView.setSieveModel(model);
                
        // Capture the Start button
        Button startButton = (Button)findViewById(R.id.startButton);
        // Register our onClick listener
        startButton.setOnClickListener(startListener);

        // Capture the thread number selector
        RadioGroup numThreadsRadioGroup = (RadioGroup)findViewById(R.id.numThreadsRadioGroup);
        // Register our onCheckedChange listener
        numThreadsRadioGroup.setOnCheckedChangeListener(numThreadsListener);
        
        // Check whether we are recreating a previously destroyed instance
        if (savedInstanceState != null) {
        	model.setNumber(savedInstanceState.getInt(KEY_NUMBER));
        	model.setDelay(savedInstanceState.getInt(KEY_DELAY));
        	model.setNumber(savedInstanceState.getInt(KEY_ALGORITHM));
        	model.setNumThreads(savedInstanceState.getInt(KEY_NUM_THREADS));
        	model.setIsPrime(savedInstanceState.getBooleanArray(KEY_IS_PRIME));
        	
        	model.forceRedraw();
        }
        
        Log.w(this.getClass().getName(), "onCreate");
    }

    // Invoked when the Activity loses user focus
    @Override
    protected void onPause() {
        super.onPause();
        
        Log.w(this.getClass().getName(), "Called onPause");
    }

    // Something is about to happen: give the Activity a chance to save state.
    @Override
    protected void onSaveInstanceState(Bundle savedInstanceState) {
    	// Save current number in our bundle
        savedInstanceState.putInt(KEY_NUMBER, model.getNumber());
        savedInstanceState.putInt(KEY_DELAY, model.getDelay());
        savedInstanceState.putInt(KEY_ALGORITHM, model.getAlgorithm());
        savedInstanceState.putInt(KEY_NUM_THREADS, model.getNumThreads());
        savedInstanceState.putBooleanArray(KEY_IS_PRIME, model.getIsPrime());
        
        // Have the View's thread save its state into our bundle
        super.onSaveInstanceState(savedInstanceState);
        
        Log.w(this.getClass().getName(), "Called onSaveInstanceState");
    }

    // Invoked during init to let the Activity set up its Menu
    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        super.onCreateOptionsMenu(menu);

        menu.add(0, MENU_FAST, 0, R.string.menu_fast);
        menu.add(0, MENU_SLOW, 0, R.string.menu_slow);
        menu.add(0, MENU_PIPELINE, 0, R.string.menu_pipeline);
        menu.add(0, MENU_RACE, 0, R.string.menu_race);
        menu.add(0, MENU_QUIT, 0, R.string.menu_quit);

        return true;
    }

    // Invoked when the user selects an item from the Menu
    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
    	int delay = 0;
    	
    	switch (item.getItemId()) {
            case MENU_FAST: // No delay
                delay = 0;
            	model.setDelay(delay);
                printStatus("Set mark delay to " + delay + "ms");
                break;
            case MENU_SLOW: // 100ms
                delay = 100;
            	model.setDelay(delay);
                printStatus("Set mark delay to " + delay + "ms");
                break;
            case MENU_PIPELINE: // Pipeline marking algorithm
            	model.setAlgorithm(SieveModel.ALGORITHM_PIPELINE);
                printStatus("Selected the Pipeline algorithm");
            	break;
            case MENU_RACE: // Race marking algorithm
            	model.setAlgorithm(SieveModel.ALGORITHM_RACE);
                printStatus("Selected the Race algorithm");
            	break;
            case MENU_QUIT: // Exit the application
            	finish();
            	break;
            default:
            	return false;
        }
        
        return true;
    }
}
