package com.android.td5;

import android.os.Handler;
import android.os.Looper;
import android.util.Log;

public class SieveModel {
	// Class for threads in the pool, using a Looper to dispatch Runnable objects
	private class MarkerThread extends Thread {
		// Handler object to queue messages to a given thread in the pool
		public Handler callbackHandler;

		@Override
		public void run() {
			Looper.prepare();
			// The callback handler to push marking requests to
			callbackHandler = new Handler();
			// Wait for marking requests indefinitely
			Looper.loop();
		}
	}
	
	// Class for the race algorithm, marking multiples of currentNumber
	private class MarkRace implements Runnable {
		@Override
		public void run() {
			// Loop until the race is over
			while (raceSeed <= currentNumber / 2) {
				// Stop immediately when the user requests another sieve
				if (numberChanged())
					break;

				// Find next (possible) prime number
				int privateSeed;
				synchronized (raceSeed) {
					privateSeed = nextSeed(raceSeed);
					raceSeed = privateSeed;
				}

				// Mark multiples of privateSeed
				markMultiples(privateSeed);
			}

			// Decrement the number of active Runnable objects in the thread pool
			synchronized (raceRunning) {
				raceRunning--;
				if (raceRunning == 0)
					countAndPrint();
			}
		}
	}

	// Class for the Pipeline algorithm, marking multiples of the constructor's argument
	private class MarkPipeline implements Runnable {
		private int seed;

		MarkPipeline(int seed) {
			this.seed = seed;
		}

		@Override
		public void run() {
			// Push the next marking operation into the pipeline
			int i = nextSeed(seed);
			if (i <= currentNumber / 2) {
				synchronized (pool) {
					currentThread = (currentThread + 1) % numThreads;
					pool[currentThread].callbackHandler.post(new MarkPipeline(i));
				}
			}

			// Mark multiples of seed
			markMultiples(seed);
			
			// If completed, compute and print the number of prime numbers
			if (i > currentNumber / 2)
				countAndPrint();
		}
	}

	// Supported algorithms
	public static final int ALGORITHM_PIPELINE = 0;
	public static final int ALGORITHM_RACE = 1;

	// Maximum number of threads in the pool
	public static final int MAX_THREADS = 4;

	// Maximum sieve upper bound
	public static final int MAX_NUMBER = 1000000;

	SieveModel(Sieve sieve) {
		this.sieve = sieve;
	}

	// Reference to the application's Activity
	private Sieve sieve;
	
	// Sieve upper bound
	private int number;

	// Protect the sieve upper bound from follow-up user requests
	private int currentNumber;
	
	// Flag to detect the need to redraw the sieve array
	private boolean primesChanged = false;

	// Number of prime integers found
	private int primes;

	// Marking delay, initially set to 0 (Fastest)
	private int delay = 0;
	
	// Upper bound on integers whose mark is visible on the application's Canvas
	private int visibleLimit;

	// Algorithm number
	private int algorithm = ALGORITHM_RACE;

	// Number of marker threads
	private int numThreads = 1;

	// Current seed for the race algorithm (Object type, for locks)
	private Integer raceSeed;

	// Termination counter for the race algorithm (Object type, for locks)
	private Integer raceRunning;

	// Index for the pipeline algorithm
	private int currentThread = 0;

	// Array to mark prime numbers
	private boolean isPrime[];

	// The pool of marker threads
	private MarkerThread pool[];

	public boolean[] getIsPrime() {
		return isPrime;
	}
	
	public void setIsPrime(boolean[] isPrime) {
		this.isPrime = isPrime;
	}
	
	public int getAlgorithm() {
		return algorithm;
	}
	
	public void setAlgorithm(int algorithm) {
		abortSieve();

		this.algorithm = algorithm;
	}
	
	public int getNumber() {
		return number;
	}
	
	public void setNumber(int number) {
		this.number = number;
	}
	
	public int getDelay() {
		return delay;
	}
	
	public void setDelay(int delay) {
		this.delay = delay;
	}

	public void setVisibleLimit(int visibleLimit) {
		this.visibleLimit = visibleLimit;
	}

	// Get the number of Marker threads
	public int getNumThreads() {
		return numThreads;
	}
	
	// Set the number of Marker threads
	public void setNumThreads(int t) {
		abortSieve();
		
		if (t < 1)
			t = 1;
		if (t > MAX_THREADS)
			t = MAX_THREADS;
		
		numThreads = t;
	}
	
	// Request the redrawing of the sieve array
	public void forceRedraw() {
		primesChanged = true;
	}
	
	// Invalidate ongoing operations
	public void abortSieve() {
		number = 0;
	}
	
	// Return true if the user triggered another sieve
	public boolean numberChanged() {
		return number != currentNumber;
	}

	// Return the update status of the sieve array and reset it to false
	public boolean needToRedraw() {
		boolean b = primesChanged;
		primesChanged = false;
		return b;
	}

	// Set the text of the status TextView
	private void printStatus(String s) {
		sieve.printStatus(s);
	}

	// Set the text of the result TextView
	private void printResult(String s) {
		sieve.printResult(s);
	}

	// Count the number of primes and display it in the result TextView
	private void countAndPrint() {
		primes = 0;

		for (int i = 2; i <= currentNumber; i++) {
			// Stop immediately when the user requests another sieve
			if (numberChanged())
				return;
			if (isPrime[i])
				primes++;
		}

		printResult("There are " + primes + " prime number"
				+ (primes != 1 ? "s" : "") + " until " + currentNumber);
		printStatus("Sieve completed");
	}
	
	// Auxiliary function to mark multiples of seed
	private void markMultiples(int seed) {
		printStatus("Marking multiples of " + seed);

		// Step by 2*seed to avoid marking even multiples (already marked as even numbers)
		int step = seed==2 ? 2 : 2*seed;
		// Start from seed*seed to avoid marking multiples of lower integers (already marked)
		int seed2 = seed * seed;
		if (seed2 < 0 || seed2 > currentNumber)
			// Note: seed2 is negative in case of arithmetic overflow
			return;
		
		// Mark odd multiples
		for (int i = seed2; i <= currentNumber; i += step)
			if (isPrime[i]) { // No need to work if already marked
				// Mark i as no n-prime
				isPrime[i] = false;

				// Trigger Canvas redraw if visible changes occurred 
				if (i < visibleLimit)
					primesChanged = true;

				// Stop immediately when the user requests another sieve
				if (numberChanged())
					return;

				// Optional delay for animation purposes
				try {
					if (delay > 0)
						Thread.sleep(delay);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
	}

	// Return the next number still marked as (possibly) prime
	private int nextSeed(int seed) {
		if (seed <= 2) // Special case for seed == 1 or seed == 2
			return seed + 1;

		// Only iterate on odd numbers
		for (int i = seed + 2; i <= currentNumber / 2; i += 2)
			if (isPrime[i])
				return i;

		// No need to look any further
		return currentNumber / 2 + 1;
	}
	
	// Initialize the pool of marker threads
	public void init() {
		pool = new MarkerThread[MAX_THREADS];
		for (int i = 0; i < MAX_THREADS; i++) {
			pool[i] = new MarkerThread();
			pool[i].start();
		}
	}
	
	// Compute the sieve of Eratosthenes
	public void sieve() {
		primes = 0; 
		// Erase result TextView
		printResult("");
		
		number = sieve.parseNumber();

		// Copy the input value to shadow variable
		// Protect the ongoing sieve from races with a new request
		// And also used to detect new requests (with a different number)
		currentNumber = number;

		// Allocate a new array
		isPrime = new boolean[currentNumber + 1];
		// Assume initially that every integer above 2 is a prime number
		for (int i = 2; i <= currentNumber; i++)
			isPrime[i] = true;

		switch (algorithm) {
		case ALGORITHM_PIPELINE:
			pool[currentThread].callbackHandler.post(new MarkPipeline(2));
			break;
		case ALGORITHM_RACE:
			raceSeed = 1;
			raceRunning = numThreads;
			for (int t = 0; t < numThreads; t++)
				pool[t].callbackHandler.post(new MarkRace());
			break;
		default:
			Log.w(this.getClass().getName(), "Unsupported algorithm selection");
		}
	}
}
