package com.android.td5;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.util.AttributeSet;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;

public class SieveSurfaceView extends SurfaceView implements SurfaceHolder.Callback {
	public class ThreadSurfaceView extends Thread {
    	ThreadSurfaceView(SurfaceHolder sieveSurfaceHolder) {
    		this.sieveSurfaceHolder = sieveSurfaceHolder;
    	}

    	// Reference to obtain and (un)lock the Surface's Canvas
    	private SurfaceHolder sieveSurfaceHolder;

    	// Whether a sieve operation is running or not
    	private boolean running = true;
 
    	// Whether it is desirable to redraw the canvas or not
    	private boolean refresh = true;
    	
    	public void setRunning(boolean running) {
    		this.running = running;
    	}

    	public void setRefresh(boolean refresh) {
    		this.refresh = refresh;
    	}
    	
    	@Override
    	public void run() {
    		while (running) {
    			// Redraw if it is desirable to refresh the display and the sieve array changed
    			boolean redraw = refresh && needToRedraw();
    			
    			if (redraw) {
    				// Access and get exclusive access to the rectangular Canvas
    				Canvas c = sieveSurfaceHolder.lockCanvas();
    				
    				eraseAll(c);
    				drawAll(c);
    				
    				// Release the rectangular Canvas
    				sieveSurfaceHolder.unlockCanvasAndPost(c);
    			}
    			
    			try {
    				// Save battery life by waking up every 40 ms only
    				Thread.sleep(40);
    			} catch (InterruptedException e) {
    				// Interrupted (it should not happen)
    			}
    		}
    	}
    }
	
	// Paint/Colors
    private final Paint TRUE_PAINT = new Paint();
    private final Paint FALSE_PAINT = new Paint();

    // Mark size
    private int size = 16;
    
    // Surface rectangle
    private int width;
    private int height;
    
	// Data and computation model of the application
    private SieveModel model;
    
    // Surface holder to access and (un)lock its canvas
	private SurfaceHolder sieveSurfaceHolder;
	
	// Thread in charge of redrawing the sieve array
	private ThreadSurfaceView threadSurfaceView;

    public SieveSurfaceView(Context context, AttributeSet attrs) {
        super(context, attrs);

        // Register our interest in hearing about changes to our surface
        sieveSurfaceHolder = getHolder();
        sieveSurfaceHolder.addCallback(this);
        
        // Initialize the Paint color of the marks in the sieve array
        TRUE_PAINT.setColor(Color.GREEN);
        FALSE_PAINT.setColor(Color.DKGRAY);
    }

    // Return the thread in charge of redrawing the sieve array
    public ThreadSurfaceView getThread() {
		return threadSurfaceView;
	}

    // Register the data and computation model of the application
	public void setSieveModel(SieveModel model) {
		this.model = model;
	}

	// Return whether the model has been updated and the Canvas needs to be redrawn
	public boolean needToRedraw() {
		return model.needToRedraw();
	}
 
	// Upper bound on integers whose mark is visible on the Canvas
	private int getVisibleLimit() {
		return width*height / (size*size);
	}

	// Round to the closest multiple of size lower than or equal to v
	private int round(int v) {
    	return (v / size) * size;
    }
 	
	// Erase the Surface's rectangular Canvas
	private void eraseAll(Canvas c) {
    	c.drawColor(Color.BLACK);
    }
	
    // Draw a mark in the rectangular Canvas at the correct position
    private void drawInteger(Canvas c, int i, boolean b) {
    	int left = (i % (width / size)) * size;
    	int top = (i / (width / size)) * size;
    	int right = left + size - 1;
    	int bottom = top + size - 1;
    	
    	c.drawRect(left, top, right, bottom, b ? TRUE_PAINT : FALSE_PAINT);
    }

    // Draw the sieve array
    private void drawAll(Canvas c) {
    	boolean isPrime[] = model.getIsPrime();
    	    	
    	if (isPrime != null) {
        	double d = Math.sqrt((double)width*height / isPrime.length);
        	size = Math.max(2, Math.min(32, (int)d));
      
    		for (int i = 0; i < isPrime.length && i < getVisibleLimit(); i++)
    			drawInteger(c, i, isPrime[i]);
    	}
    }
 
    // The Surface has been created and is ready to be used
    @Override
    public void surfaceCreated(SurfaceHolder holder) {
    	// Note that surfaceChanged is called later on
    	
        threadSurfaceView = new ThreadSurfaceView(sieveSurfaceHolder);
        threadSurfaceView.start();
        
        Log.w(this.getClass().getName(), "Called surfaceCreated");
    }

    // The Surface/Canvas has been destroyed; it must never be touched again
    @Override
    public void surfaceDestroyed(SurfaceHolder holder) {
    	// Tell thread to shut down, and wait for it to finish, or else
        // it might touch the SurfaceView after the function returns
        boolean retry = true;
        threadSurfaceView.setRunning(false);

        while (retry) {
            try {
                threadSurfaceView.join();
                retry = false;
            } catch (InterruptedException e) {
            	// Retry
            }
        }

        Log.w(this.getClass().getName(), "Called surfaceDestroyed");
    }

    // Pause on focus lost. e.g., user switches to take a call
    @Override
    public void onWindowFocusChanged(boolean hasWindowFocus) {
    	threadSurfaceView.setRefresh(hasWindowFocus);
    	
        Log.w(this.getClass().getName(), "Called onWindowFocusChanged");
    }

    // The surface dimensions change
    @Override
    public void surfaceChanged(SurfaceHolder holder, int format, int width, int height) {
    	this.width = round(width);
    	this.height = round(height);

    	model.setVisibleLimit(getVisibleLimit());

        Log.w(this.getClass().getName(), "Called surfaceChanged");
    }
}
